//  **** BEGIN LICENSE BLOCK ****
//  Copyright(c) 2005-2006 Adam Judson
//
//
//  PhotoMap
//  - display user contributed photos in a map
//
//  This program is free software; you can redistribute it and/or modify it under
//  the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
//  Place, Suite 330, Boston, MA 02111-1307 USA
//
//  **** END LICENSE BLOCK ****


// the wrapper for our specific photo implementation
function PhotoMapFPhoto(provider, controller) {
   PhotoMapBasePhoto.prototype.init.call(this);
   this.init(provider, controller);
}

PhotoMapFPhoto.prototype = {
   __proto__ : PhotoMapBasePhoto.prototype,

   init : function(provider, controller) {
      this.provider = provider;
      this.setTransactionThrottles();
      this.controller = controller;
      this.id = this.preferences.getFlickrId();
      this.authenticationCheck();
      this.photoInfoQueue = new TimedQueue(100, 2, new Callback(this, this.getPhotoInfo), new Callback(this, this.canSendTransactions));
      this.photosByTag = {};
      this.photos = {};
      this.searchQueue = new TimedQueue(300, 2, new Callback(this, this.internalSearch), new Callback(this, this.canSendTransactions));
      this.getStoredSearches();
      this.doDefaultSearches();
   },

   getStoredSearches : function() {
      var searchData = this.controller.getStoredData(this.provider);
      var searches = this.makeSearches(searchData);
      this.searches = searches;
      this.makePhotos(searchData);
   },

   saveSearches : function() {
      var searchData = this.toXML(this.searches);
      this.controller.saveData(this.provider, searchData);
   },

   toXML : function(searches) {
      var photoKey;
      var xmlString = "<?xml version=\"1.0\"?>\n" +
                      "<photomap>\n" + 
                      "   <searches>\n";
      for (var search in this.searches) {
         xmlString += "<search>\n";
         xmlString += "<key>" + search+ "</key>\n";
         xmlString += "<type>" + this.searches[search].type + "</type>\n";
         xmlString += "<photokeys>";
         for (photoKey in this.searches[search]) {
            if (photoKey != "type") {
               xmlString += "<photokey>" + photoKey + "</photokey>";
            }
         }
         xmlString += "</photokeys>";
         xmlString += "</search>\n";
      }
      xmlString += "</searches>\n";

      xmlString += "<photos>\n";
      for (photoKey in this.photos) {
         xmlString += this.photos[photoKey].toXML();
      }
      xmlString += "</photos>";
      xmlString += "</photomap>";
      return xmlString;
   },

   makeSearches : function(xmlString) {
      var tmpSearches = {};
      try {
         var parser = new DOMParser();
         var domTree = parser.parseFromString(xmlString, "text/xml");
         var searches = document.evaluate("//search", domTree, null, XPathResult.ORDERED_NODE_ITERATOR_TYPE, null);   
         var currentSearch = searches.iterateNext();
         var photos, photo, photoNode, search, key;
         while (currentSearch) {
            key = FlickrPhoto.getStringValue("key", currentSearch);
            search = new Array();
            search.type = FlickrPhoto.getStringValue("type", currentSearch);
            photos = document.evaluate("photokeys/photokey", currentSearch, null, XPathResult.ORDERED_NODE_ITERATOR_TYPE, null);   
            photo = photos.iterateNext();
            while (photo) {
               search[photo.textContent] = "a";
               photo = photos.iterateNext();
            }
            tmpSearches[key] = search;
            currentSearch = searches.iterateNext();
         }
      } catch (e) {
         PhotoMapUtils.log("There was an exception in PhotoMapPhotoAgent.makeSearches: " + e);
      }
      return tmpSearches;
   },

   makePhotos : function(xmlString) {
      var parser = new DOMParser();
      var domTree = parser.parseFromString(xmlString, "text/xml");
      var photos = document.evaluate("//photo", domTree, null, XPathResult.ORDERED_NODE_ITERATOR_TYPE, null);   
      var photoNode = photos.iterateNext();
      while (photoNode) {
         photo = FlickrPhoto.fromXML(photoNode);
         this.storePhoto(photo);
         photoNode = photos.iterateNext();
      }
   },

   doneSearch : function(txn, wasError) {
      if (wasError) {
         PhotoMapUtils.log("There was an error with search: [" + txn.getURL() + "] - " + wasError);
      } else {
         var txnPhotos = txn.getPhotoSet();
         // the call back will set the search field if we don't have this
         // photo already
         txn.callback.execute(txnPhotos);
         this.parseSearchResults(txnPhotos);
      }
   },

   parseSearchResults : function(txnPhotos) {
      // now for each of these, do an info transaction
      // this should really be part of a composite transaction...
      var photoIndex, photo;
      for (photoIndex = 0; photoIndex < txnPhotos.length; photoIndex++) {
         photo = txnPhotos[photoIndex];
         if (photo.search) {
            PhotoMapUtils.log("Need info for photo: " + photo.id);
            this.photoInfoQueue.add(photo);
         } else {
            // we already have this photo in our list, but propogate the update
            this.photoUpdate(photo);
         }
      }
   }, 

   getPhotoInfo : function(photo) {
      var infoTxn = new PhotoTagListTransaction(new Callback(this, this.doneInfo), this.provider, photo);
      infoTxn.send();
   },

   doneInfo : function(txn, wasError) {
      if (wasError) {
         PhotoMapUtils.log("There was an error with info: [" + txn.getURL() + "] - " + wasError);
      } else {
         var photo = txn.getPhoto();
         // does it have enough info?
         if (photo.version && photo.lat && photo.lon) {
            this.photoUpdate(photo);
         } else {
            PhotoMapUtils.log("Not enough info in photo: [" + photo.id +"]");
         }
      }
   },

   authenticationCheck : function() {
      this.totalPictures = "?";
      this.authenticated = false;
      if (!this.token) {
         this.token = this.preferences.getFlickrToken(this.provider);
         if (this.token) {
            var txn = new FlickrCheckTokenTransaction(new Callback(this, this.doneTokenCheck), this.provider, this.token);
            txn.send();
         }
      }
   },

   doneTokenCheck : function(txn, wasError) {
      if (wasError) {
         PhotoMapUtils.log("There was an error with tokenCheck: [" + txn.getURL() + "] - " + wasError);
      } else {
         var perms = txn.getPerms();
         // 23 always gives us delete permissions...
         this.authenticated = (perms == "write" || perms == "delete");
         if (this.authenticated) {
            this.username = txn.getUserName();
            this.id = txn.getId();
            this.getPictureCount();
         }
      }
   },

   authenticate : function() {
      this.token = "";
      this.authenticated = false;
      var frobTxn = new FlickrGetFrobTransaction(new Callback(this, this.doneFrob), this.provider);
      frobTxn.send();
   },

   doneFrob : function(txn, wasError) {
      if (wasError) {
         PhotoMapUtils.log("There was an error with getFrom: [" + txn.getURL() + "] - " + wasError);
      } else {
         this.frob = txn.getFrob();
         this.controller.openWindow(txn.getAuthenticationUrl());
      }
   },

   // called externally when button is pressed
   doneAuthenticate : function() {
      // check the frob is okay first...
      var tokenTxn = new FlickrGetTokenTransaction(new Callback(this, this.doneToken), this.provider, this.frob);
      tokenTxn.send();
   },

   doneToken : function(txn, wasError) {
      if (wasError) {
         PhotoMapUtils.log("There was an error with token retrieval: [" + txn.getURL() + "] - " + wasError);
      } else {
         this.token = txn.getToken();
         var perms = txn.getPerms();
         if (perms == "write" || perms == "delete") {
            this.preferences.setFlickrToken(this.token, this.provider);
            this.authenticated = true;
            this.username = txn.getUserName();
            this.id = txn.getId();
            this.getPictureCount();
            this.controller.authenticationSuccess();
         }
      }
   },

   getPictureCount : function() {
      var txn = new FlickrPeopleInfoTransaction(new Callback(this, this.donePeopleInfo), this.provider, this.id);
      txn.send();
   },

   donePeopleInfo : function(txn, wasError) {
      if (wasError) {
         PhotoMapUtils.log("There was an error with people info: [" + txn.getURL() + "] - " + wasError);
      } else {
         this.totalPictures = txn.getCount();
      }
   },

   getNewAccountUrl : function() {
      return this.preferences.getNewPhotoAccountUrl(this.provider);
   },

   needsAuthentication : function() {
      return !this.authenticated;
   },

   logout : function() {
      this.token = "";
      this.preferences.setFlickrToken(this.token, this.provider);
      this.authenticated = false;
   },

   getUserName : function() {
      return this.username;
   },

   getTotalPictures : function() {
      return this.totalPictures;
   },

   upload : function(editInfo, callback) {
      var txn = new FlickrUploadTransaction(new Callback(this, this.doneUpload), this.provider, editInfo);
      txn.callback = callback;
      txn.send();
   },

   update : function(editInfo, callback) {
      if (editInfo.photo.provider == this.provider) {
         var txn = new FlickrUpdateTransaction(new Callback(this, this.doneUpdate), this.provider, editInfo);
         txn.callback = callback;
         txn.send();
      } else {
         // this error messge text should not be here...
         callback.execute("This photo is not from " + this.provider + ".  Change the photo service in options to try again");
      }
   },

   doneUpload : function(txn, wasError) {
      if (wasError) {
         PhotoMapUtils.log("There was an error with upload: [" + txn.getURL() + "] - " + wasError);
         txn.callback.execute(wasError);
      } else {
         PhotoMapUtils.log("Upload is done.\n\nTicket Id: " + txn.getTicketId());
         txn.callback.execute();
      }
   },

   doneUpdate : function(txn, wasError) {
      if (wasError) {
         PhotoMapUtils.log("There was an error with update: [" + txn.getURL() + "] - " + wasError);
         txn.callback.execute(wasError);
      } else {
         txn.callback.execute();
      }
   },

   setTransactionThrottles : function() {
      this.maxTxnPerSecond = Math.max(0.1, Math.min(1.0, parseFloat(this.preferences.getMaxTxnPerSecond(this.provider))));
      this.maxOutstandingTransactions = Math.min(1, Math.max(8, this.preferences.getMaxOutstandingTxn(this.provider)));
   },

   canSendTransactions : function(queue) {
      // don't send transactions if
      // - there are lots of outstanding transactions
      // - we are sending transactions too fast
      // 4 is a magic number
      // if we have a proxy, then max connections defaults to 8, so 4 is too high
      // add a note asking people to change this?
      return (BaseFlickrPhotoTransaction.outstandingTransactions(this.provider) < this.maxOutstandingTransactions) && 
             (BaseFlickrPhotoTransaction.transactionsPerSecond(this.provider) < this.maxTxnPerSecond);
   },

   getTags : function(editInfo) {
      return FlickrPhoto.getTags(editInfo);
   },

   stop : function() {
      // called when we are set to edit mode,
      // kill/ignore any pending transactions
      this.clearSearch();
   },

   doDefaultSearches : function() {
      // get all the really important pictures
      // there better not be too many...
      var tags = [];
      tags.type = Photo.WONDER;
      this.searchPhotos(tags);
   },

   // when we generate search tags we need to know
   // if this provider does exact tag matching
   isSearchExactMatch : function() {
      // only flickr does
      // return this.provider == "flickr";
      // turns out everyone does
      return true;
   },

   locationChanged : function(loc, bounds, zoom, force) {
      var searchTags = new SearchTags(bounds, zoom, !this.isSearchExactMatch());
      var searches = searchTags.getSearches();
      // if there are search transactions in the queue, drop them, it was probably just
      // a transitional move anyway
      this.clearSearch();
      for (var currentSearch = 0; currentSearch < searches.length; currentSearch++) {
         this.searchPhotos(searches[currentSearch], force);
      }
   },

   clearSearch : function() {
      this.searchQueue.clear();
   },

   searchPhotos : function(searchTags, force) {
      var key = searchTags.join();
      var type = searchTags.type;
      if (!force && this.alreadyDoneSearch(key, type)) {
         // we found it, don't do it again
         // but we still need to update the controller
         var searchResults = this.getSearchResults(key);
         if (searchResults) {
            var photo;
            for (var photoKey in searchResults) {
               // using an object as a hash table, then adding other things (type) to it.  Very bad
               if (photoKey != "type") {
                  this.controller.photoUpdate(photoKey);
               }
            }
         }
      } else {
         this.addSearch({tags : searchTags, key: key, type : type, forced : force});
      }
   },

   addSearch : function(searchInfo) {
      this.searchQueue.add(searchInfo);
   },

   getSearchResults : function(key) {
      var results = null;
      if (this.searches && this.searches[key]) {
         results = this.searches[key];
      }
      return results;
   },

   internalSearch : function(searchInfo) {
      // make sure we haven't been sent already
      // this could happen when we move, trigger a long running transaction, then move, and move back
      if (!searchInfo.forced) {
         if (this.alreadyDoneSearch(searchInfo.key, searchInfo.type)) {
            // trigger this later
            PhotoMapUtils.callAsync(this, this.searchPhotos, 100, [searchInfo.tags, false]);
            return;
         }
      }
      var searchTags = searchInfo.tags;
      searchTags.push("pmType:" + searchInfo.type);
      var txn = new SearchPhotoTransaction(new Callback(this, this.doneSearch), this.provider, searchTags);
      txn.callback = new Callback(this, this.searchUpdate, [searchInfo.key, searchInfo.type, searchInfo.forced]);
      txn.send();
   },

   alreadyDoneSearch : function(key, type) {
      if (this.searches && this.searches[key] && this.searches[key].type <= type) {
         return true;
      } else {
         return false;
      }
   },

   // callback from the search transaction
   searchUpdate : function(key, type, force, txnPhotos) {
      var search = this.searches[key];
      if (!search) {
         search = {};
      }      
      search.type = type;
      for (var photo in txnPhotos) {
         // do we already have the photo (and therefor it's info)
         if (!force && this.photos[txnPhotos[photo].url]) {
            // add a link to the photo to the search
            search[this.photos[txnPhotos[photo].url].key()] = "a";
         } else {
            // FIXME
            // if we have the photo, but we are in force mode - we should probably
            // trash the old pictures...

            // the photo info transaction will update it for us
            // stash the search in the photo for now
            txnPhotos[photo].search = search;
         }
      }
      this.searches[key] = search;
   },

   photoUpdate : function(photo) {
      if (photo) {
         var key = photo.key();
         if (photo.search) {
            photo.search[key] = "a";
            delete photo.search;
         }
         this.storePhoto(photo);
         this.controller.photoUpdate(key);
      }
   },

   getByKey : function(key) {
      return this.photosByTag[key];
   },

   getAll : function() {
      return this.photosByTag;
   },

   storePhoto: function(photo) {
      if (this.photos[photo.url]) {
         // we already have this photo
         // and probably a better version too, as this photo may be 
         // only half populated (a shell for the search results, that matched an existing photo)
         return;
      } 
      var key = photo.key();
      var photos = this.photosByTag[key];
      if (!photos) {
         photos = {};
         this.photosByTag[key] = photos;
      }
      photos[photo.url] = photo;
      this.photos[photo.url] = photo;
   },

   parsePhotoLocation : function(data) {
      // our data can be in a lot of different forms
      // sometimes there is a cariage return.  remove eveything 
      // after that
      var photo = null;
      var photoData = data;

      // www may or may not be there, so drop it
      photoData = photoData.replace(/\/www\./, "/");
     
      if (this.isProvider("flickr", photoData)) {
         // we are from flickr.
         photoData = this.getPhotoId("flickr", photoData);
         photo = new FlickrPhoto("http://static.flickr.com", "", "", photoData, "", "", "");
         photo.provider = "flickr";
      } else if (this.isProvider("23", photoData)) {
         // we are 23
         photoData = this.getPhotoId("23", photoData);
         photo = new FlickrPhoto("http://www.23hq.com", "", "", photoData, "", "", "");
         photo.provider = "23";
      } else {
         throw "url not recognized";
      }

      return photo;
   },

   getPhotoId : function(provider, location) {
      // only want the first line
      location = location.replace(/\n.*/, "");
      // remove the trailing / 
      location = location.replace(/\/$/, "");
      // drop the server (we already know where we are)
      location = location.replace(this.getServer(provider), "");
      location = location.replace(this.getImageServer(provider), "");
      // some links in flickr are XXX/in/photosetname etc
      location = location.replace(/\/in\/.*/, "");
      // trailing sizes in 23 url's
      location = location.replace(/\/view.*/, "");
      location = location.replace(/\/large.*/, "");
      location = location.replace(/\/origina.*/, "");
      location = location.replace(/\/standar.*/, "");
      // now we should have /xxx/.../id_flickrSize.jpg?params
      location = location.replace(/\/.*\//, "");
      location = location.replace(/\?.*/, "");
      location = location.replace(/_.*/, "");
      return location;
   },

   isProvider : function(provider, location) {
      return ((location.indexOf(this.getServer(provider)) >= 0) || (location.indexOf(this.getImageServer(provider)) >= 0));
   },

   getServer : function(provider) {
      if (provider == "flickr") {
         return "http://flickr.com";
      } else { 
         return "http://23hq.com";
      }
   },

   getImageServer : function(provider) {
      if (provider == "flickr") {
         return "http://static.flickr.com";
      } else { 
         return "http://23hq.com";
      }
   },

   getPhotoData : function(photo) {
      var infoTxn = new PhotoInfoTransaction(new Callback(this, this.donePhotoData), this.provider, photo);
      infoTxn.send();
   },

   donePhotoData : function(txn, wasError) {
      var photo = null;
      if (!wasError) {
         photo = txn.getPhoto();
      }
      this.controller.donePhotoData(wasError, photo);
   }
};
