
/* components defined in this file */
const PHOTOMAP_CONTENT_HANDLER_CONTRACTID =
    "@mozilla.org/uriloader/content-handler;1?type=x-application-photomap";
const PHOTOMAP_CONTENT_HANDLER_CID =
    Components.ID("{9c51bd27-6ed8-4000-a2bf-36cb95c0c997}");

const PHOTOMAP_PROTOCOL_HANDLER_CONTRACTID =
    "@mozilla.org/network/protocol;1?name=photomap";
const PHOTOMAP_PROTOCOL_HANDLER_CID =
    Components.ID("{9c51bd27-6ed8-4000-a2bf-36cb95c0c998}");

/* components used by this file */
const STANDARDURL_CONTRACTID =
    "@mozilla.org/network/standard-url;1";

const IOSERVICE_CONTRACTID = 
    "@mozilla.org/network/io-service;1";

/* interafces used in this file */
const nsIContentHandler  = Components.interfaces.nsIContentHandler;
const nsISupports        = Components.interfaces.nsISupports;
const nsIURI             = Components.interfaces.nsIURI;
const nsIStandardURL     = Components.interfaces.nsIStandardURL;
const nsIChannel         = Components.interfaces.nsIChannel;
const nsIProtocolHandler = Components.interfaces.nsIProtocolHandler;
const nsIRequest         = Components.interfaces.nsIRequest;
const nsIIOService       = Components.interfaces.nsIIOService;



/* x-application-photomap content handler */
function PhotoMapContentHandler() {
}

PhotoMapContentHandler.prototype.QueryInterface = function(iid) {
   if (!iid.equals(nsIContentHandler)) {
      throw Components.results.NS_ERROR_NO_INTERFACE;
   }
   return this;
};

PhotoMapContentHandler.prototype.handleContent = function (contentType, windowTarget, request, aRemovedArg) {
   // do nothing...
};


/* content handler factory object (PhotoMapContentHandler) */
var PhotoMapContentHandlerFactory = new Object();

PhotoMapContentHandlerFactory.createInstance = function(outer, iid) {
   if (outer != null) {
      throw Components.results.NS_ERROR_NO_AGGREGATION;
   }

   if (!iid.equals(nsIContentHandler) && !iid.equals(nsISupports)) {
      throw Components.results.NS_ERROR_INVALID_ARG;
   }
   return new PhotoMapContentHandler();
};

/* PhotoMap protocol handler component */
function PhotoMapProtocolHandler() {}

PhotoMapProtocolHandler.prototype.scheme = "photomap";

PhotoMapProtocolHandler.prototype.defaultPort = 4100;

PhotoMapProtocolHandler.prototype.protocolFlags = nsIProtocolHandler.URI_NORELATIVE | nsIProtocolHandler.ALLOWS_PROXY;

PhotoMapProtocolHandler.prototype.allowPort = function (port, scheme) { 
   return false; 
};

PhotoMapProtocolHandler.prototype.newURI = function(spec, charset, baseURI) {
   var cls = Components.classes[STANDARDURL_CONTRACTID];
   var url = cls.createInstance(nsIStandardURL);
   url.init(nsIStandardURL.URLTYPE_STANDARD, 4100, spec, charset, baseURI);
   return url.QueryInterface(nsIURI);
};

PhotoMapProtocolHandler.prototype.newChannel = function(URI) {
   var ios = Components.classes[IOSERVICE_CONTRACTID].getService(nsIIOService);
   if (!ios.allowPort(URI.port, URI.scheme)) {
      throw Components.results.NS_ERROR_FAILURE;
   }

   var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService();
   var windowManagerInterface = windowManager.QueryInterface(Components.interfaces.nsIWindowMediator);
   var topWindow = windowManagerInterface.getMostRecentWindow('global:PhotoMap');
   var uri = URI.spec;

   var loc = {};

   var paramstring = uri.substring(uri.indexOf("?") + 1, uri.length);
   paramstring = paramstring.replace(/&amp;/g, "&");
   var params = paramstring.split("&");
   var nameValue;
   for (var p in params) {
      nameValue = params[p].split("=");
      if (nameValue[0] == "lat") {
         loc.lat = parseFloat(nameValue[1]);
      } else if (nameValue[0] == "long") {
         loc.lon = parseFloat(nameValue[1]);
      } else if (nameValue[0] == "zoom") {
         loc.zoom = parseFloat(nameValue[1]);
      } else if (nameValue[0] == "type") {
         loc.type = nameValue[1];
      }

   }

   if (topWindow) {
      topWindow.setTimeout(function() {topWindow.PhotoMapGUI.instance.controller.setLocation(loc);topWindow.focus();}, 1000);
      // topWindow.focus();
   } else {
      var w = windowManager.getMostRecentWindow("navigator:browser");
      if (w) {
         // pass arguments...
         w.openDialog('chrome://photomap/content/PhotoMap.xul', "_blank", "chrome,extrachrome,menubar,resizable,scrollbars,status,toolbar", loc);
      }
   }
   return new PhotoMapChannel(URI);
};

/* protocol handler factory object (PhotoMapProtocolHandler) */
var PhotoMapProtocolHandlerFactory = new Object();

PhotoMapProtocolHandlerFactory.createInstance = function (outer, iid) {
   if (outer != null) {
      throw Components.results.NS_ERROR_NO_AGGREGATION;
   }
   if (!iid.equals(nsIProtocolHandler) && !iid.equals(nsISupports)) {
      throw Components.results.NS_ERROR_INVALID_ARG;
   }
   return new PhotoMapProtocolHandler();
};

/* bogus channel used by the PhotoMapProtocolHandler */
function PhotoMapChannel(URI) {
   this.URI = URI;
   this.originalURI = URI;
}

PhotoMapChannel.prototype.QueryInterface = function (iid) {
   if (!iid.equals(nsIChannel) && !iid.equals(nsIRequest) && !iid.equals(nsISupports)) {
      throw Components.results.NS_ERROR_NO_INTERFACE;
   }
   return this;
};

/* nsIChannel */
PhotoMapChannel.prototype.loadAttributes = null;
PhotoMapChannel.prototype.contentType = "x-application-photomap";
PhotoMapChannel.prototype.contentLength = 0;
PhotoMapChannel.prototype.owner = null;
PhotoMapChannel.prototype.loadGroup = null;
PhotoMapChannel.prototype.notificationCallbacks = null;
PhotoMapChannel.prototype.securityInfo = null;

PhotoMapChannel.prototype.open = function() { 
   throw Components.results.NS_ERROR_NOT_IMPLEMENTED; 
};

PhotoMapChannel.prototype.asyncOpen = function(observer, ctxt) {
   observer.onStartRequest(this, ctxt);
};

PhotoMapChannel.prototype.asyncRead = function(listener, ctxt) {
   return listener.onStartRequest (this, ctxt);
};

/* nsIRequest */
PhotoMapChannel.prototype.isPending = function() { 
   return true; 
};

PhotoMapChannel.prototype.status = Components.results.NS_OK;

PhotoMapChannel.prototype.cancel = function(status) {
   this.status = status;
};

PhotoMapChannel.prototype.suspend = function() {
   throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
};

PhotoMapChannel.prototype.resume = function() {
   throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
};

/*****************************************************************************/

var PhotoMapModule = new Object();

PhotoMapModule.registerSelf = function (cm, fileSpec, location, type) {
    cm = cm.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    cm.registerFactoryLocation(PHOTOMAP_CONTENT_HANDLER_CID,
                               "PhotoMap content handler",
                               PHOTOMAP_CONTENT_HANDLER_CONTRACTID, 
                               fileSpec, 
                               location,
                               type);
    cm.registerFactoryLocation(PHOTOMAP_PROTOCOL_HANDLER_CID,
                               "PhotoMap protocol handler",
                               PHOTOMAP_PROTOCOL_HANDLER_CONTRACTID, 
                               fileSpec, 
                               location,
                               type);
};

PhotoMapModule.unregisterSelf = function(cm, fileSpec, location) {
    cm = cm.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    cm.unregisterFactoryLocation(PHOTOMAP_CONTENT_HANDLER_CID, fileSpec);
    cm.unregisterFactoryLocation(PHOTOMAP_PROTOCOL_HANDLER_CID, fileSpec);
};

PhotoMapModule.getClassObject = function (cm, cid, iid) {
   if (cid.equals(PHOTOMAP_CONTENT_HANDLER_CID)) {
      return PhotoMapContentHandlerFactory;
   } else if (cid.equals(PHOTOMAP_PROTOCOL_HANDLER_CID)) {
      return PhotoMapProtocolHandlerFactory;
   } else if (!iid.equals(Components.interfaces.nsIFactory)) {
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
   } else {
      throw Components.results.NS_ERROR_NO_INTERFACE;
   }
};

PhotoMapModule.canUnload = function(cm) { 
   return true; 
};

/* entrypoint */
function NSGetModule(cm, fileSpec) { 
   return PhotoMapModule; 
}
